TorsionInvariants := module()
option package;
export 

##########################################################################################
# License: Apache License, Version 2.0
# Copyright [2023] [Joshua James Leiter]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
##########################################################################################

#Update 10/25/2022
#Created a TorsionInvariants Module for the calculation of these in regards to the Torsion Omega spinor Classification.
#End Update: 10/25/2022

#Update 10/26/2022
#Added several more in(co)variants.
# TorsionOmegaCovariant   #Zero in Cases 8
# TorsionNuCovariant   #Zero in Cases 7,8
# TorsionUpsilonCovariant #Zero in Cases 4,6,7,8
# TorsionThetaCovariant     #Zero in Cases 5,6,7,8
# TorsionI6Invariant #Zero in Cases 4,5,6,7,8
# TorsionRCovariant #Zero in Cases 3,4,5,6,7,8
# TorsionDeltaCovariant #Zero in Cases 2,3,4,5,6,7,8
#
# TorsionIpInvariant #For the Delta Covariant
# TorsionIpInvariant #For the Delta Covariant
#
#Also added an algorithm for the omega spinor: TorsionOmegaSpinorReducibilityAlgorithm
#End Update: 10/26/2022

#Update 10/26/2022
#Added several more in(co)variants for the Rank 2 spinors which appear in the decomp. of the torsion tensor.
# TorsionRank2AlphaCovariant   #Zero in Cases 3
# TorsionRank2OmicronInvariant   #Zero in Cases 2
#
#Also added an algorithm for the xi and theta spinors: TorsionRank2
#End Update: 10/26/2022

#Update 11/1/2022
#Added return options for TorsionOmegaSpinorReducibilityAlgorithm & TorsionRank2ReducibilityAlgorithm instead of print.
#Added TorsionTensorReducibilityAlgorithm (14), which classifies the torsion tensor based on spinor reducibility.
#End Update: 11/1/2022

#Update 2/16/2023
#Updated The torsion omega algorithm to be more clear and match the dissertation.
#End Update 2/16/2023

#Update 2/25/2023
#Updated to include quotes in some of the LHS pieces in the lists
#End Update 2/25/2023

#Update 3/7/2023
#Updated TorsionOmegaSpinorReducibilityAlgorithm to remove redundant Theta calculation.
#End Update 3/7/2023

#Update 3/9/2023
#Updated TorsionOmegaSpinorReducibilityAlgorithm to update & include notation from dissertation.
#Updated TorsionRank2ReducibilityAlgorithm to have aSolderForm as an argument, which it should.
#End Update 3/9/2023

#Below are the Torsion invariants I created. They have a similar flavor to Zakhary and Carminati. These are in Commands 1-14

### Numbers 1-10 are for the Torsion Omega Spinor

#(*1*)
	TorsionOmegaCovariant,
#(*2*)
	TorsionNuCovariant,
#(*3*)
	TorsionUpsilonCovariant,
#(*4*)
	TorsionThetaCovariant,
#(*5*)
	TorsionI6Invariant,
#(*6*)
	TorsionRCovariant,
#(*7*)
	TorsionIpInvariant,
#(*8*)
	TorsionJpInvariant,
#(*9*)
	TorsionDeltaInvariant,
#(*10*)
	TorsionOmegaSpinorReducibilityAlgorithm,

#Numbers 11-13 are for the Torsion Xi and Theta Spinors. (It also classifies a general rank 2 hermitian spinor: alpha(AA') 

#(*11*)
	TorsionRank2AlphaCovariant,
#(*12*)
	TorsionRank2OmicronInvariant,
#(*13*)
	TorsionRank2ReducibilityAlgorithm,
#(*14*)
	TorsionTensorReducibilityAlgorithm;
#(*15*)

#(*16*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*1*)

TorsionOmegaCovariant:=proc(aOmegaSpinor)
local Dz1,Dz2,Dw1,Dw2,_Omega0,_Omega1,_Omega2,_Omega3,_Omega4,_Omega5,_Omega6,_Omega7,
Omega0,Omega1,Omega2,Omega3,Omega4,Omega5,Omega6,Omega7,myList;

#This command takes a spinor from SSSS*
if DGinfo(aOmegaSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "This command takes a spinor from SSSS*: invalid spinor type, expecting type  ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aOmegaSpinor)

fi;

Dz1:=DGinfo("FrameFiberVectors")[1];
Dz2:=DGinfo("FrameFiberVectors")[2];
Dw1:=DGinfo("FrameFiberVectors")[3];
Dw2:=DGinfo("FrameFiberVectors")[4];

#Build the Omega Scalars:
_Omega0:=simplify(Hook([Dz1,Dz1,Dz1,Dw1],aOmegaSpinor));
_Omega1:=simplify(Hook([Dz1,Dz1,Dz1,Dw2],aOmegaSpinor));
_Omega2:=simplify(Hook([Dz1,Dz1,Dz2,Dw1],aOmegaSpinor));
_Omega3:=simplify(Hook([Dz1,Dz1,Dz2,Dw2],aOmegaSpinor));
_Omega4:=simplify(Hook([Dz1,Dz2,Dz2,Dw1],aOmegaSpinor));
_Omega5:=simplify(Hook([Dz1,Dz2,Dz2,Dw2],aOmegaSpinor));
_Omega6:=simplify(Hook([Dz2,Dz2,Dz2,Dw1],aOmegaSpinor));
_Omega7:=simplify(Hook([Dz2,Dz2,Dz2,Dw2],aOmegaSpinor));

myList:=[
Omega0=_Omega0,
Omega1=_Omega1,
Omega2=_Omega2,
Omega3=_Omega3,
Omega4=_Omega4,
Omega5=_Omega5,
Omega6=_Omega6,
Omega7=_Omega7];

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*2*)

TorsionNuCovariant:=proc(aPhiSpinor)
local myList1,myEvalList,myNuList,Omega0,Omega1,Omega2,Omega3,Omega4,Omega5,Omega6,Omega7,Nu0,Nu1,Nu2,Nu3,Nu4,Nu5,Nu6,Nu7,Nu8;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=TorsionOmegaCovariant(aPhiSpinor);

#Build the Omega Scalars:
myEvalList:=[
Omega0=rhs(myList1[1]),
Omega1=rhs(myList1[2]),
Omega2=rhs(myList1[3]),
Omega3=rhs(myList1[4]),
Omega4=rhs(myList1[5]),
Omega5=rhs(myList1[6]),
Omega6=rhs(myList1[7]),
Omega7=rhs(myList1[8])];

#Build the Nu Covariant:
myNuList:=simplify(eval([
Nu0=2*Omega0*Omega4-2*Omega2^2,
Nu1=2*Omega0*Omega5+2*Omega1*Omega4-4*Omega2*Omega3,
Nu2=2*Omega1*Omega5-2*Omega3^2,
Nu3=2*Omega0*Omega6-2*Omega2*Omega4,
Nu4=2*Omega0*Omega7+2*Omega1*Omega6-2*Omega2*Omega5-2*Omega3*Omega4,
Nu5=2*Omega1*Omega7-2*Omega3*Omega5,
Nu6=2*Omega2*Omega6-2*Omega4^2,
Nu7=2*Omega2*Omega7+2*Omega3*Omega6-4*Omega4*Omega5,
Nu8=2*Omega3*Omega7-2*Omega5^2],myEvalList));

return myNuList;
end proc:


#1111111111111111111111111111111111111111111111111111111111111111111111
#(*3*)

TorsionUpsilonCovariant:=proc(aPhiSpinor)
local myList1,myEvalList,Nu0,Nu1,Nu2,Nu3,Nu4,Nu5,Nu6,Nu7,Nu8,Upsilon0,Upsilon1,Upsilon2,Upsilon3,Upsilon4,myUpsilonList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=TorsionNuCovariant(aPhiSpinor);

#Build the Nu Scalars:
myEvalList:=[
Nu0=rhs(myList1[1]),
Nu1=rhs(myList1[2]),
Nu2=rhs(myList1[3]),
Nu3=rhs(myList1[4]),
Nu4=rhs(myList1[5]),
Nu5=rhs(myList1[6]),
Nu6=rhs(myList1[7]),
Nu7=rhs(myList1[8]),
Nu8=rhs(myList1[9])];

#Build the Upsilon Covariant:
myUpsilonList:=simplify(eval([
Upsilon0=2*Nu0*Nu6-(1/2)*Nu3^2,
Upsilon1=2*Nu0*Nu7+2*Nu1*Nu6-Nu3*Nu4,
Upsilon2=2*Nu0*Nu8-Nu3*Nu5+2*Nu2*Nu6+2*Nu1*Nu7-(1/2)*Nu4^2,
Upsilon3=2*Nu1*Nu8+2*Nu2*Nu7-Nu4*Nu5,
Upsilon4=2*Nu2*Nu8-(1/2)*Nu5^2
],myEvalList));

return myUpsilonList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*4*)

TorsionThetaCovariant:=proc(aPhiSpinor)
local myList1,myEvalList,Omega0,Omega1,Omega2,Omega3,Omega4,Omega5,Omega6,Omega7,
Theta0,Theta1,Theta2,Theta3,Theta4,myThetaList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=TorsionOmegaCovariant(aPhiSpinor);

#Build the Omega Scalars:
myEvalList:=[
Omega0=rhs(myList1[1]),
Omega1=rhs(myList1[2]),
Omega2=rhs(myList1[3]),
Omega3=rhs(myList1[4]),
Omega4=rhs(myList1[5]),
Omega5=rhs(myList1[6]),
Omega6=rhs(myList1[7]),
Omega7=rhs(myList1[8])];

#Build the Theta Covariant:
myThetaList:=simplify(eval([
Theta0=2*Omega0*Omega3-2*Omega1*Omega2,
Theta1=4*Omega0*Omega5-4*Omega1*Omega4,
Theta2=2*Omega0*Omega7-2*Omega1*Omega6+6*Omega2*Omega5-6*Omega3*Omega4,
Theta3=4*Omega2*Omega7-4*Omega3*Omega6,
Theta4=2*Omega4*Omega7-2*Omega5*Omega6],myEvalList));

return myThetaList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*5*)

TorsionI6Invariant:=proc(aOmegaSpinor)
local myList1,myEvalList,Omega0,Omega1,Omega2,Omega3,Omega4,Omega5,Omega6,Omega7,I6,myI6;

#This command takes a spinor from SSSS*
if DGinfo(aOmegaSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "This command takes a spinor from SSSS*: invalid spinor type, expecting type  ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aOmegaSpinor)

fi;

#Get the Omega Scalars:
myList1:=TorsionOmegaCovariant(aOmegaSpinor);

#Build the Omega Scalars:
myEvalList:=[
Omega0=rhs(myList1[1]),
Omega1=rhs(myList1[2]),
Omega2=rhs(myList1[3]),
Omega3=rhs(myList1[4]),
Omega4=rhs(myList1[5]),
Omega5=rhs(myList1[6]),
Omega6=rhs(myList1[7]),
Omega7=rhs(myList1[8])];

#Build the I6 Invariant:
myI6:=[I6=simplify(eval(2*Omega0*Omega7-2*Omega1*Omega6-6*Omega2*Omega5+6*Omega3*Omega4,myEvalList))];

return myI6;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*6*)

TorsionRCovariant:=proc(aPhiSpinor)
local myList1,myEvalList,Theta0,Theta1,Theta2,Theta3,Theta4,R0,R1,R2,R3,R4,R5,R6,myRList;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=TorsionThetaCovariant(aPhiSpinor);

#Build the Theta Scalars:
myEvalList:=[
Theta0=rhs(myList1[1]),
Theta1=rhs(myList1[2]),
Theta2=rhs(myList1[3]),
Theta3=rhs(myList1[4]),
Theta4=rhs(myList1[5])];

#Build the R Covariant:
myRList:=simplify(eval([
R0=-8*Theta0^2*Theta3+4*Theta0*Theta1*Theta2-Theta1^3,
R1=-32*Theta0^2*Theta4-4*Theta0*Theta1*Theta3+8*Theta0*Theta2^2-2*Theta1^2*Theta2,
R2=-40*Theta0*Theta1*Theta4+20*Theta0*Theta2*Theta3-5*Theta1^2*Theta3,
R3=20*Theta0*Theta3^2-20*Theta1^2*Theta4,
R4=40*Theta0*Theta3*Theta4-20*Theta1*Theta2*Theta4+5*Theta1*Theta3^2,
R5=32*Theta0*Theta4^2+4*Theta1*Theta3*Theta4-8*Theta2^2*Theta4+2*Theta2*Theta3^2,
R6=8*Theta1*Theta4^2-4*Theta2*Theta3*Theta4+Theta3^3],myEvalList));

return myRList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*7*)

TorsionIpInvariant:=proc(aPhiSpinor)
local myList1,myEvalList,Theta0,Theta1,Theta2,Theta3,Theta4,Ip,myIp;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=TorsionThetaCovariant(aPhiSpinor);

#Build the Theta Scalars:
myEvalList:=[
Theta0=rhs(myList1[1]),
Theta1=rhs(myList1[2]),
Theta2=rhs(myList1[3]),
Theta3=rhs(myList1[4]),
Theta4=rhs(myList1[5])];

#Build the R Covariant:
myIp:=simplify(eval([Ip = 2*Theta0*Theta4-(1/2)*Theta1*Theta3+(1/6)*Theta2^2],myEvalList));

return myIp;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*8*)

TorsionJpInvariant:=proc(aPhiSpinor)
local myList1,myEvalList,Theta0,Theta1,Theta2,Theta3,Theta4,Jp,myJp;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

myList1:=TorsionThetaCovariant(aPhiSpinor);

#Build the Theta Scalars:
myEvalList:=[
Theta0=rhs(myList1[1]),
Theta1=rhs(myList1[2]),
Theta2=rhs(myList1[3]),
Theta3=rhs(myList1[4]),
Theta4=rhs(myList1[5])];

#Build the R Covariant:
myJp:=simplify(eval([Jp = Theta0*Theta2*Theta4-(1/36)*Theta2^3-(3/8)*Theta0*Theta3^2-(3/8)*Theta1^2*Theta4+(1/8)*Theta3*Theta1*Theta2],myEvalList));

return myJp;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*9*)

TorsionDeltaInvariant:=proc(aPhiSpinor)
local myIp,myJp,myI,myJ,myEvalList,myDelta;

if DGinfo(aPhiSpinor, "TensorIndexType") <> ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

error "invalid spinor type, expecting type ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aPhiSpinor)

fi;

#Build the Ip and Jp Invariants
myIp:=TorsionIpInvariant(aPhiSpinor);
myJp:=TorsionJpInvariant(aPhiSpinor);

myEvalList:=[myI=rhs(op(myIp)),myJ=rhs(op(myJp))];

#Build the Delta Covariant:
myDelta:=simplify(eval([Delta = myI^3-(6)*myJ^2],myEvalList));

return myDelta;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*10*)

TorsionOmegaSpinorReducibilityAlgorithm:=proc(aTorsionTensorOrSpinor,aSolderForm,myOmegaSpinor:=0)
local OmegaSpinor,
Omega,OmegaSol,
Nu,NuSol,
Upsilon,UpsilonSol,
Theta,ThetaSol,
R,RSol,
Delta,DeltaSol;

#####Error messages & Splitting

###One: If Recieved Torsion tensor
#This little tidbit here is if you want to do the classification straight from a torsion tensor & solder form.
if DGinfo(aTorsionTensorOrSpinor, "TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	
	OmegaSpinor:=TorsionOmegaSpinor(aTorsionTensorOrSpinor,aSolderForm);

###Two: If Recieved Omega Spinor
elif DGinfo(aTorsionTensorOrSpinor, "TensorIndexType") = ["cov_vrt", "cov_vrt", "cov_vrt", "cov_vrt"] then

	OmegaSpinor:=aTorsionTensorOrSpinor;

###Three: If not recieved either
else

error "invalid tensor or spinor type, expecting either type ['cov_bas', 'cov_bas', 'cov_bas'] or ['cov_vrt', 'cov_vrt', 'cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aTorsionTensorOrSpinor)

fi;


######Now we run the algorithm we have developed above.

###
# Step 1: Calculate the Omega Covariant
###

Omega:=TorsionOmegaCovariant(OmegaSpinor);
OmegaSol:=convert(map(rhs,Omega),set);

#if OmegaSol={0} then Case 8 else calculate Nu.
if OmegaSol={0} then 
	print("Case 8, [-]-(-), the zero spinor");
	return "(-)";
fi;

###
# Step 2: Calculate the Nu Covariant
###

Nu:=TorsionNuCovariant(OmegaSpinor);
NuSol:=convert(map(rhs,Nu),set);

#if NuSol={0} then Case 7 else calculate Upsilon.
if NuSol={0} then 
	print("Case 7, [A^3,B']-(1,0)^3(0,1), delta & delta & delta & xi");
	return "(1,0)^3(0,1)";
fi;

###
# Step 3: Calculate the Upsilon and Theta Covariants
###

Upsilon:=TorsionUpsilonCovariant(OmegaSpinor);
UpsilonSol:=convert(map(rhs,Upsilon),set);
Theta:=TorsionThetaCovariant(OmegaSpinor);
ThetaSol:=convert(map(rhs,Theta),set);



#if UpsilonSol={0} then either Case 4 or Case 6 else check Theta
if UpsilonSol={0} then 	
	if ThetaSol={0} then 
		print("Case 6, [A^2,B,C']-(1,0)^2(1,0)(0,1), delta & delta & omega & xi");
		return "(1,0)^2(1,0)(0,1)";
	else 
		print("Case 4, [A^2,11]-(1,0)^2(1,1), delta & delta & theta");
		return "(1,0)^2(1,1)";
	fi;
fi;

#If Upsilon non zero then cannot be case 4,6,7,8. We check Theta
#if ThetaSol={0} then Case 5 else calculate R
if ThetaSol={0} then 
	print("Case 5, [A,B,C,D']-(1,0)(1,0)(1,0)(0,1), delta & lambda &t omega & xi");
	return "(1,0)(1,0)(1,0)(0,1)";
fi;


###
# Step 4: Calculate the R Covariant
###

R:=TorsionRCovariant(OmegaSpinor);
RSol:=convert(map(rhs,R),set);

#if RSol={0} then Case 3 else calculate Delta
if RSol={0} then 
print("Case 3, [A,B,11]-(1,0)(1,0)(1,1), delta & lambda & theta");
return "(1,0)(1,0)(1,1)";
fi;

###
# Step 5: Calculate the Delta/Discriminant Invariant
###

Delta:=TorsionDeltaInvariant(OmegaSpinor);
DeltaSol:=convert(map(rhs,Delta),set);

#if DeltaSol={0} then Case 2 else Case 1
if DeltaSol={0} then 
print("Case 2, [A,21]-(1,0)(2,1), delta & beta");
return;
else
print("Case 1, [31]-(3,1), Omega irreducible");
return "(3,1)";
fi;

end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*11*)

TorsionRank2AlphaCovariant:=proc(aRank2HermitianSpinor)
local myFiberBasis,Dz1,Dz2,Dw1,Dw2,myalpha00,myalpha01,myalpha10,myalpha11,myAlphaList;

myFiberBasis:=DGinfo("FrameFiberVectors");
Dz1:=myFiberBasis[1];
Dz2:=myFiberBasis[2];
Dw1:=myFiberBasis[3];
Dw2:=myFiberBasis[4];

myalpha00:=Hook([Dz1,Dw1],aRank2HermitianSpinor);
myalpha01:=Hook([Dz1,Dw2],aRank2HermitianSpinor);
myalpha10:=Hook([Dz2,Dw1],aRank2HermitianSpinor);
myalpha11:=Hook([Dz2,Dw2],aRank2HermitianSpinor);

myAlphaList:=[alpha00=myalpha00,alpha01=myalpha01,alpha10=myalpha10,alpha11=myalpha11];

return myAlphaList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*12*)

TorsionRank2OmicronInvariant:=proc(aRank2HermitianSpinor)
local myEvalList,myOmicronList;

myEvalList:=TorsionRank2AlphaCovariant(aRank2HermitianSpinor);

myOmicronList:=simplify(eval([Omicron=2*alpha00*alpha11-2*alpha01*alpha10],myEvalList));

return myOmicronList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*13*)

TorsionRank2ReducibilityAlgorithm:=proc(aTorsionTensorOrSpinor,aSolderForm,o:=0)
local mySpinor,alpha,alphaSol,Omicron,OmicronSol;

#####Error messages & Splitting

###One: If Recieved Torsion tensor
#This little tidbit here is if you want to do the classification straight from a torsion tensor & solder form.
if DGinfo(aTorsionTensorOrSpinor, "TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then
	
	if o="Xi" then
	mySpinor:=TorsionXiSpinor(aTorsionTensorOrSpinor,aSolderForm);

	elif o="Theta" then
	mySpinor:=TorsionThetaSpinor(aTorsionTensorOrSpinor,aSolderForm);

	fi;

###Two: If Recieved Xi or Theta Spinor
elif DGinfo(aTorsionTensorOrSpinor, "TensorIndexType") = ["cov_vrt", "cov_vrt"] then

	mySpinor:=aTorsionTensorOrSpinor;

###Three: If not recieved either
else

error "invalid tensor or spinor type, expecting either type ['cov_bas', 'cov_bas', 'cov_bas'] or ['cov_vrt', 'cov_vrt'], Recieved:", DGErrorPrint(aTorsionTensorOrSpinor)

fi;

###
# Step 1: Calculate the alpha Covariant
###

#Calculate alpha:
alpha:=TorsionRank2AlphaCovariant(mySpinor);
alphaSol:=convert(map(rhs,alpha),set);

#if alphaSol={0} then Case 3 else calculate Omicron
if alphaSol={0} then 
print("Case 3, (-), the zero spinor");
return "(-)";
fi;

###
# Step 2: Calculate the Omicron Covariant
###

Omicron:=TorsionRank2OmicronInvariant(mySpinor);
OmicronSol:=convert(map(rhs,Omicron),set);

#if OmicronSol={0} then Case 2 else Case 1
if OmicronSol={0} then 
print("Case 2, |(1,0)|^2, xi & xiBar");
return "|(1,0)|^2";
else
print("Case 1, (1,1), spinor irreducible");
return "(1,1)";
fi;

end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*14*)

TorsionTensorReducibilityAlgorithm:=proc(aTorsionTensor,aSolder)
local g,myTorsionTensor,myThetaSpinor,myXiSpinor,myOmegaSpinor,ThetaType,XiType,OmegaType,Theta,Xi,Omega,myList;

g:=SpinorInnerProduct(aSolder,aSolder);

###Here we add a piece to classify the Torsion tensor if it comes with a contravariant index

if DGinfo(aTorsionTensor,"TensorIndexType") = ["con_bas", "cov_bas", "cov_bas"] then
	
	myTorsionTensor:=RaiseLowerIndices(g,aTorsionTensor,[1]);

elif DGinfo(aTorsionTensor,"TensorIndexType") = ["cov_bas", "cov_bas", "cov_bas"] then

	myTorsionTensor:=aTorsionTensor;

else

	error "invalid tensor type, expecting type ['con_bas', 'cov_bas', 'cov_bas'] or ['cov_bas', 'cov_bas', 'cov_bas'], Recieved:", DGErrorPrint(aTorsionTensor)

fi;

###Now we move onto the classification:

#Get the SL(2,C) irreducible spinors
myThetaSpinor:=TorsionThetaSpinor(myTorsionTensor,aSolder);
myXiSpinor:=TorsionXiSpinor(myTorsionTensor,aSolder);
myOmegaSpinor:=TorsionOmegaSpinor(myTorsionTensor,aSolder);

#Determine the spinor structure reducibility
ThetaType:=TorsionRank2ReducibilityAlgorithm(myThetaSpinor);
XiType:=TorsionRank2ReducibilityAlgorithm(myXiSpinor);
OmegaType:=TorsionOmegaSpinorReducibilityAlgorithm(myOmegaSpinor);

#Output a list of all the reducibility types:
myList:=[Theta=ThetaType,Xi=XiType,Omega=OmegaType];

return myList;
end proc:

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*15*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#(*16*)

#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111
#1111111111111111111111111111111111111111111111111111111111111111111111

end module: